"""

Nature constants, in SI units.

Source: https://physics.nist.gov/cuu/Constants/index.html 

"""

# Exact values
e_0         =       1.602176634e-19         # [C]   Elementary charge
h           =       6.62607015e-34          # [JS]  Planck constant, exact
k_b         =       1.380649e-23            # [J/K] Boltzmann constant, exact

# Measured values
m_e         =       9.1093837015e-31        # [kg]  Electron mass
m_p         =       1.67262192369e-27       # [kg]  Proton mass
h_bar       =       1.054571817e-34         # [Js]  Reduced Planck constant  
𝜀_0         =       8.8541878128e-12        # [F/m] Vacuum permittivity
